@@ -21,7 +21,7 @@ function wesnoth.micro_ais.recruit_rushers(cfg)
 	local optional_keys = { "high_level_fraction", "randomness" }
 	local CA_parms = {
 		ai_id = 'mai_rusher_recruit',
-		{ ca_id = "move", location = 'ca_recruit_rushers.lua', score = cfg.ca_score or 180000 }
+		{ ca_id = "move", location = '../../lua/ca_recruit_rushers.lua', score = cfg.ca_score or 180000 }
 	}
 
 	handle_default_recruitment(cfg)
