@@ -14,6 +14,7 @@ map_extensions   = ("map", "mask")
 image_extensions = ("png", "jpg", "jpeg")
 sound_extensions = ("ogg", "wav")
 vc_directories = (".git", ".svn")
+misc_files_extensions = ("-bak", ".DS_Store", "Thumbs.db") # These files and extensions should be included in the `default_blacklist` in filesystem.hpp.
 l10n_directories = ("l10n",)
 resource_extensions = map_extensions + image_extensions + sound_extensions
 image_reference = r"[A-Za-z0-9{}.][A-Za-z0-9_/+{}.\-\[\]~\*,]*\.(png|jpe?g)(?=(~.*)?)"
@@ -67,8 +68,16 @@ if no expansion could be performed"""
                 match_range = re.match("(\d+)~(\d+)", token) # range syntax, eg [1~4]
                 if match_range:
                     before, after = int(match_range.group(1)), int(match_range.group(2))
+                    # does one of the limits have leading zeros? If so, detect the length of the numbers used
+                    if match_range.group(1).startswith("0") or match_range.group(2).startswith("0"):
+                        leading_zeros = max(len(match_range.group(1)), len(match_range.group(2)))
+                    else:
+                        leading_zeros = 0
                     incr = 1 if before <= after else -1 # to allow iterating in reversed order, eg. [4~1]
-                    substitutions[i].extend([Substitution(str(n), match.start(0), match.end(0)) for n in range(before, after + incr, incr)])
+                    # previously this code used a mere casting to str
+                    # string formatting allows proper handling of leading zeros, if any
+                    fmt_string = "{:0" + str(leading_zeros) + "d}"
+                    substitutions[i].extend([Substitution(fmt_string.format(n), match.start(0), match.end(0)) for n in range(before, after + incr, incr)])
                     continue
                 substitutions[i].append(Substitution(token, match.start(0), match.end(0))) # no operator found
 
@@ -81,7 +90,8 @@ if no expansion could be performed"""
         for sub_array in substitutions:
             new_string = path
             for sub in reversed(sub_array): # to avoid creating "holes" in the strings
-                yield new_string[:sub.start] + sub.sub + new_string[sub.end:] # these are the expanded strings
+                new_string = new_string[:sub.start] + sub.sub + new_string[sub.end:] # these are the expanded strings
+            yield new_string
 
 def is_root(dirname):
     "Is the specified path the filesystem root?"
@@ -147,13 +157,12 @@ def parse_attribute(line):
     where = line.find("=")
     leader = line[:where]
     after = line[where+1:]
-    after = after.lstrip()
     if re.search("\s#", after):
         where = len(re.split("\s+#", after)[0])
-        value = after[:where]
+        value = after[:where].lstrip()
         comment = after[where:]
     else:
-        value = after.rstrip()
+        value = after.strip()
         comment = ""
     # Return four fields: stripped key, part of line before value,
     # value, trailing whitespace and comment.
@@ -215,15 +224,16 @@ class Forest:
             rest = [elem for elem in subtree if not elem.endswith("_main.cfg")]
             subtree = sorted(maincfgs) + sorted(rest)
             self.forest.append(subtree)
-        for i in self.forest:
-            # Ignore version-control subdirectories and Emacs tempfiles
+        for i in range(len(self.forest)):
+            # Ignore version-control subdirectories, Emacs tempfiles, and other files and extensions
             for dirkind in vc_directories + l10n_directories:
-                i = [x for x in i if dirkind not in x]
-            i = [x for x in i if '.#' not in x]
-            i = [x for x in i if not os.path.isdir(x)]
+                self.forest[i] = [x for x in self.forest[i] if dirkind not in x]
+            self.forest[i] = [x for x in self.forest[i] if '.#' not in x]
+            self.forest[i] = [x for x in self.forest[i] if not os.path.isdir(x)]
             if exclude:
-                i = [x for x in i if not re.search(exclude, x)]
-            i = [x for x in i if not x.endswith("-bak")]
+                self.forest[i] = [x for x in self.forest[i] if not re.search(exclude, x)]
+            for each in misc_files_extensions:
+                self.forest[i] = [x for x in self.forest[i] if not x.endswith(each)]
         # Compute cliques (will be used later for visibility checks)
         self.clique = {}
         counter = 0
@@ -278,9 +288,33 @@ def isresource(filename):
     return ext and ext[1:] in resource_extensions
 
 def parse_macroref(start, line):
+    def handle_argument():
+        nonlocal opt_arg
+        nonlocal arg
+        nonlocal optional_args
+        nonlocal args
+
+        arg = arg.strip()
+        # is this an optional argument?
+        # argument names are usually made of uppercase letters, numbers and underscores
+        # if they're optional, they're followed by an equal sign
+        # stop matching on the first one, because the argument value might contain one too
+        if re.match(r"^([A-Z0-9_]+?)=", arg):
+            opt_arg, arg = arg.split("=", 1)
+        if arg.startswith('"') and arg.endswith('"'):
+            arg = arg[1:-1].strip()
+        if opt_arg:
+            optional_args[opt_arg] = arg
+            opt_arg = ""
+        else:
+            args.append(arg)
+        arg = ""
+
     brackdepth = parendepth = 0
     instring = False
+    optional_args = {}
     args = []
+    opt_arg = ""
     arg = ""
     for i in range(start, len(line)):
         if instring:
@@ -298,11 +332,7 @@ def parse_macroref(start, line):
             brackdepth -= 1
             if brackdepth == 0:
                 if not line[i-1].isspace():
-                    arg = arg.strip()
-                    if arg.startswith('"') and arg.endswith('"'):
-                        arg = arg[1:-1].strip()
-                    args.append(arg)
-                    arg = ""
+                    handle_argument()
                 break
             else:
                 arg += line[i]
@@ -314,14 +344,10 @@ def parse_macroref(start, line):
              line[i].isspace() and \
              brackdepth == 1 and \
              parendepth == 0:
-            arg = arg.strip()
-            if arg.startswith('"') and arg.endswith('"'):
-                arg = arg[1:-1].strip()
-            args.append(arg)
-            arg = ""
+            handle_argument()
         elif not line[i].isspace() or parendepth > 0:
             arg += line[i]
-    return (args, brackdepth, parendepth)
+    return (args, optional_args, brackdepth, parendepth)
 
 def formaltype(f):
     # Deduce the expected type of the formal
@@ -410,10 +436,18 @@ def actualtype(a):
         atype = "string"
     return atype
 
-def argmatch(formals, actuals):
+def argmatch(formals, optional_formals, actuals, optional_actuals):
+    if optional_formals:
+        for key in optional_actuals.keys():
+            if key not in optional_formals:
+                return False
     if len(formals) != len(actuals):
         return False
-    for (f, a) in zip(formals, actuals):
+    opt_formals, opt_actuals = [], []
+    for key, value in optional_actuals.items():
+        opt_formals.append(key)
+        opt_actuals.append(value)
+    for (f, a) in zip(formals + opt_formals, actuals + opt_actuals):
         # Here's the compatibility logic.  First, we catch the situations
         # in which a more restricted actual type matches a more general
         # formal one.  Then we have a fallback rule checking for type
@@ -456,17 +490,22 @@ def argmatch(formals, actuals):
 @total_ordering
 class Reference:
     "Describes a location by file and line."
-    def __init__(self, namespace, filename, lineno=None, docstring=None, args=None):
+    def __init__(self, namespace, filename, lineno=None, docstring=None, args=None,
+                 optional_args=None, deprecated=False, deprecation_level=0, removal_version=None):
         self.namespace = namespace
         self.filename = filename
         self.lineno = lineno
         self.docstring = docstring
         self.args = args
+        self.optional_args = optional_args
+        self.deprecated = deprecated
+        self.deprecation_level = deprecation_level
+        self.removal_version = removal_version
         self.references = collections.defaultdict(list)
         self.undef = None
 
-    def append(self, fn, n, a=None):
-        self.references[fn].append((n, a))
+    def append(self, fn, n, args=None, optional_args=None):
+        self.references[fn].append((n, args, optional_args))
 
     def dump_references(self):
         "Dump all known references to this definition."
@@ -486,10 +525,10 @@ class Reference:
             return self.filename > other.filename
 
     def mismatches(self):
-        copy = Reference(self.namespace, self.filename, self.lineno, self.docstring, self.args)
+        copy = Reference(self.namespace, self.filename, self.lineno, self.docstring, self.args, self.optional_args)
         copy.undef = self.undef
         for filename in self.references:
-            mis = [(ln,a) for (ln,a) in self.references[filename] if a is not None and not argmatch(self.args, a)]
+            mis = [(ln,a,oa) for (ln,a,oa) in self.references[filename] if a is not None and not argmatch(self.args, self.optional_args, a, oa)]
             if mis:
                 copy.references[filename] = mis
         return copy
@@ -547,6 +586,8 @@ class CrossRef:
     def scan_for_definitions(self, namespace, filename):
         ignoreflag = False
         conditionalsflag = False
+        temp_docstrings = {}
+        current_docstring = None
         with codecs.open(filename, "r", "utf8") as dfp:
             state = "outside"
             latch_unit = in_base_unit = in_theme = False
@@ -608,11 +649,39 @@ class CrossRef:
                               % (filename, n+1), file=sys.stderr)
                     else:
                         name = tokens[1]
-                        here = Reference(namespace, filename, n+1, line, args=tokens[2:])
+                        here = Reference(namespace, filename, n+1, line, args=tokens[2:], optional_args=[])
                         here.hash = hashlib.md5()
                         here.docstring = line.lstrip()[8:] # Strip off #define_
+                        current_docstring = None
+                        if name in temp_docstrings:
+                            here.docstring += temp_docstrings[name]
+                            del temp_docstrings[name]
                         state = "macro_header"
                     continue
+                if state in ('outside', 'external_docstring'):
+                    # allow starting new docstrings even one after another
+                    m = re.match(r"\s*# wmlscope: docstring (\w+)", line)
+                    if m:
+                        current_docstring = m.group(1)
+                        # what if someone tries to define a docstring twice for the same macro?
+                        # In this case, warn and overwrite the old one
+                        if current_docstring in temp_docstrings:
+                            print("Redefining a docstring for macro {} at {}, line {}".format(m.group(1),
+                                                                                              filename,
+                                                                                              n+1),
+                                  file=sys.stderr)
+                        temp_docstrings[current_docstring] = ""
+                        state = "external_docstring"
+                        continue
+                if state == 'external_docstring':
+                    # stop collecting the docstring on the first non-comment line (even if it's empty)
+                    # if the line starts with a #define or another docstring directive
+                    # it'll be handled in the blocks above
+                    if line.lstrip().startswith("#"):
+                        temp_docstrings[current_docstring] += line.lstrip()[1:]
+                    else:
+                        current_docstring = None
+                        state = 'outside'
                 elif state != 'outside' and line.strip().endswith("#enddef"):
                     here.hash.update(line.encode("utf8"))
                     here.hash = here.hash.digest()
@@ -632,13 +701,38 @@ class CrossRef:
                         self.xref[name] = []
                     self.xref[name].append(here)
                     state = "outside"
-                elif state == "macro_header" and line.strip() and line.strip()[0] != "#":
-                    state = "macro_body"
+                elif state == "macro_header" and line.strip():
+                    if line.strip().startswith("#arg"):
+                        state = "macro_optional_argument"
+                        here.optional_args.append(line.strip().split()[1])
+                    elif line.strip()[0] != "#":
+                        state = "macro_body"
+                elif state == "macro_optional_argument" and "#endarg" in line:
+                    state = "macro_header"
+                    continue
                 if state == "macro_header":
                     # Ignore macro header commends that are pragmas
-                    if "wmlscope" not in line and "wmllint:" not in line:
+                    if ("wmlscope" in line) or ("wmllint:" in line):
+                        continue
+                    # handle deprecated macros
+                    if "deprecated" in line:
+                        # There are four levels of macro deprecation (1, 2, 3 and 4)
+                        # Sometimes they have a version number in which they're
+                        # scheduled for removal and sometimes they don't have it
+                        # This regex seems to match every deprecated macro in mainline
+                        # in version 1.15.6
+                        m = re.match(r"\s*#\s?deprecated\s(1|2|3|4)\s?([0-9.]*)\s?(.*)", line)
+                        if m:
+                            here.deprecated = True
+                            # leave them as strings: they'll be used for HTML output
+                            here.deprecation_level = m.group(1)
+                            here.removal_version = m.group(2)
+                            here.docstring += m.group(3)
+                        else:
+                            print("Deprecation line not matched found in {}, line {}".format(filename, n+1), file=sys.stderr)
+                    else:
                         here.docstring += line.lstrip()[1:]
-                if state in ("macro_header", "macro_body"):
+                if state in ("macro_header", "macro_optional_argument", "macro_body"):
                     here.hash.update(line.encode("utf8"))
                 elif line.strip().startswith("#undef"):
                     tokens = line.split()
@@ -669,6 +763,15 @@ class CrossRef:
                                 self.unit_ids[uid] = []
                             self.unit_ids[uid].append(Reference(namespace, filename, n+1))
                             latch_unit= False
+        # handling of the file is over, but there are some external docstring still around
+        # this happens if someone defined an external docstring *after* the macro it refers to
+        # or to a macro which isn't even in the file
+        # warn if that's the case
+        if temp_docstrings: # non-empty dictionaries cast as True
+            print("Docstrings defined after their macros or referring to a missing \
+macro found in {}: {}".format(filename,
+                              ", ".join(temp_docstrings.keys())),
+                  file=sys.stderr)
     def __init__(self, dirpath=[], exclude="", warnlevel=0, progress=False):
         "Build cross-reference object from the specified filelist."
         self.filelist = Forest(dirpath, exclude)
@@ -702,7 +805,9 @@ class CrossRef:
         # Next, decorate definitions with all references from the filelist.
         self.unresolved = []
         self.missing = []
+        self.deprecated = []
         formals = []
+        optional_formals = []
         state = "outside"
         if self.warnlevel >=2 or progress:
             print("*** Beginning reference-gathering pass...")
@@ -715,11 +820,17 @@ class CrossRef:
                     have_icon = False
                     beneath = 0
                     ignoreflag = False
+                    in_macro_definition = False
                     for (n, line) in enumerate(rfp):
                         if line.strip().startswith("#define"):
                             formals = line.strip().split()[2:]
+                            in_macro_definition = True
                         elif line.startswith("#enddef"):
                             formals = []
+                            optional_formals = []
+                            in_macro_definition = False
+                        elif in_macro_definition and line.startswith("#arg"):
+                            optional_formals.append(line.strip().split()[1])
                         comment = ""
                         if '#' in line:
                             if "# wmlscope: start ignoring" in line:
@@ -751,15 +862,16 @@ class CrossRef:
                             if self.warnlevel >=2:
                                 print('"%s", line %d: seeking definition of %s' \
                                       % (fn, n+1, name))
-                            if name in formals:
+                            if name in formals or name in optional_formals:
                                 continue
                             elif name in self.xref:
                                 # Count the number of actual arguments.
                                 # Set args to None if the call doesn't
                                 # close on this line
-                                (args, brackdepth, parendepth) = parse_macroref(match.start(0), line)
+                                (args, optional_args, brackdepth, parendepth) = parse_macroref(match.start(0), line)
                                 if brackdepth > 0 or parendepth > 0:
                                     args = None
+                                    optional_args = None
                                 else:
                                     args.pop(0)
                                 #if args:
@@ -768,8 +880,10 @@ class CrossRef:
                                 # Figure out which macros might resolve this
                                 for defn in self.xref[name]:
                                     if self.visible_from(defn, fn, n+1):
-                                        defn.append(fn, n+1, args)
+                                        defn.append(fn, n+1, args, optional_args)
                                         candidates.append(str(defn))
+                                        if defn.deprecated:
+                                            self.deprecated.append((name,Reference(ns,fn,n+1)))
                                 if len(candidates) > 1:
                                     print("%s: more than one definition of %s is visible here (%s)." % (Reference(ns, fn, n), name, "; ".join(candidates)))
                             if len(candidates) == 0:
@@ -798,7 +912,7 @@ class CrossRef:
                                     # it as a reference to everything the substitutions
                                     # could potentially match.
                                     elif '{' in name or '@' in name:
-                                        pattern = re.sub(r"(\{[^}]*\}|@R0|@V)", '.*', name)
+                                        pattern = re.sub(r"(\{[^}]*\}|@R[0-5]|@V)", '.*', name)
                                         key = self.mark_matching_resources(pattern, fn,n+1)
                                         if key:
                                             self.fileref[key].append(fn, n+1)
